@@ -1,30 +1,48 @@
 package org.jsoup;
 
 import org.jsoup.nodes.Document;
+import org.jsoup.parser.HtmlTreeBuilder;
 import org.jsoup.parser.Parser;
 
-import java.net.URL;
-import java.util.Map;
-import java.util.Collection;
 import java.io.IOException;
+import java.io.InputStream;
+import java.net.URL;
+import java.util.Collection;
+import java.util.Map;
 
 /**
  * A Connection provides a convenient interface to fetch content from the web, and parse them into Documents.
  * <p>
  * To get a new Connection, use {@link org.jsoup.Jsoup#connect(String)}. Connections contain {@link Connection.Request}
  * and {@link Connection.Response} objects. The request objects are reusable as prototype requests.
+ * </p>
  * <p>
  * Request configuration can be made using either the shortcut methods in Connection (e.g. {@link #userAgent(String)}),
- * or by methods in the Connection.Request object directly. All request configuration must be made before the request
- * is executed.
+ * or by methods in the Connection.Request object directly. All request configuration must be made before the request is
+ * executed.
+ * </p>
  */
 public interface Connection {
 
     /**
      * GET and POST http methods.
      */
-    public enum Method {
-        GET, POST
+    enum Method {
+        GET(false), POST(true), PUT(true), DELETE(false), PATCH(true);
+
+        private final boolean hasBody;
+
+        Method(boolean hasBody) {
+            this.hasBody = hasBody;
+        }
+
+        /**
+         * Check if this HTTP method has/needs a request body
+         * @return if body needed
+         */
+        public final boolean hasBody() {
+            return hasBody;
+        }
     }
 
     /**
@@ -32,21 +50,21 @@ public interface Connection {
      * @param url URL to connect to
      * @return this Connection, for chaining
      */
-    public Connection url(URL url);
+    Connection url(URL url);
 
     /**
      * Set the request URL to fetch. The protocol must be HTTP or HTTPS.
      * @param url URL to connect to
      * @return this Connection, for chaining
      */
-    public Connection url(String url);
+    Connection url(String url);
 
     /**
      * Set the request user-agent header.
      * @param userAgent user-agent to use
      * @return this Connection, for chaining
      */
-    public Connection userAgent(String userAgent);
+    Connection userAgent(String userAgent);
 
     /**
      * Set the request timeouts (connect and read). If a timeout occurs, an IOException will be thrown. The default
@@ -54,7 +72,7 @@ public interface Connection {
      * @param millis number of milliseconds (thousandths of a second) before timing out connects or reads.
      * @return this Connection, for chaining
      */
-    public Connection timeout(int millis);
+    Connection timeout(int millis);
 
     /**
      * Set the maximum bytes to read from the (uncompressed) connection into the body, before the connection is closed,
@@ -63,28 +81,28 @@ public interface Connection {
      * @param bytes number of bytes to read from the input before truncating
      * @return this Connection, for chaining
      */
-    public Connection maxBodySize(int bytes);
+    Connection maxBodySize(int bytes);
 
     /**
      * Set the request referrer (aka "referer") header.
      * @param referrer referrer to use
      * @return this Connection, for chaining
      */
-    public Connection referrer(String referrer);
+    Connection referrer(String referrer);
 
     /**
      * Configures the connection to (not) follow server redirects. By default this is <b>true</b>.
      * @param followRedirects true if server redirects should be followed.
      * @return this Connection, for chaining
      */
-    public Connection followRedirects(boolean followRedirects);
+    Connection followRedirects(boolean followRedirects);
 
     /**
      * Set the request method to use, GET or POST. Default is GET.
      * @param method HTTP request method
      * @return this Connection, for chaining
      */
-    public Connection method(Method method);
+    Connection method(Method method);
 
     /**
      * Configures the connection to not throw exceptions when a HTTP error occurs. (4xx - 5xx, e.g. 404 or 500). By
@@ -93,7 +111,7 @@ public interface Connection {
      * @param ignoreHttpErrors - false (default) if HTTP errors should be ignored.
      * @return this Connection, for chaining
      */
-    public Connection ignoreHttpErrors(boolean ignoreHttpErrors);
+    Connection ignoreHttpErrors(boolean ignoreHttpErrors);
 
     /**
      * Ignore the document's Content-Type when parsing the response. By default this is <b>false</b>, an unrecognised
@@ -103,39 +121,70 @@ public interface Connection {
      * Document.
      * @return this Connection, for chaining
      */
-    public Connection ignoreContentType(boolean ignoreContentType);
+    Connection ignoreContentType(boolean ignoreContentType);
 
     /**
-     * Add a request data parameter. Request parameters are sent in the request query string for GETs, and in the request
-     * body for POSTs. A request may have multiple values of the same name.
+     * Disable/enable TSL certificates validation for HTTPS requests.
+     * <p>
+     * By default this is <b>true</b>; all
+     * connections over HTTPS perform normal validation of certificates, and will abort requests if the provided
+     * certificate does not validate.
+     * </p>
+     * <p>
+     * Some servers use expired, self-generated certificates; or your JDK may not
+     * support SNI hosts. In which case, you may want to enable this setting.
+     * </p>
+     * <p>
+     * <b>Be careful</b> and understand why you need to disable these validations.
+     * </p>
+     * @param value if should validate TSL (SSL) certificates. <b>true</b> by default.
+     * @return this Connection, for chaining
+     */
+    Connection validateTLSCertificates(boolean value);
+
+    /**
+     * Add a request data parameter. Request parameters are sent in the request query string for GETs, and in the
+     * request body for POSTs. A request may have multiple values of the same name.
      * @param key data key
      * @param value data value
      * @return this Connection, for chaining
      */
-    public Connection data(String key, String value);
+    Connection data(String key, String value);
+
+    /**
+     * Add an input stream as a request data paramater. For GETs, has no effect, but for POSTS this will upload the
+     * input stream.
+     * @param key data key (form item name)
+     * @param filename the name of the file to present to the remove server. Typically just the name, not path,
+     * component.
+     * @param inputStream the input stream to upload, that you probably obtained from a {@link java.io.FileInputStream}.
+     * You must close the InputStream in a {@code finally} block.
+     * @return this Connections, for chaining
+     */
+    Connection data(String key, String filename, InputStream inputStream);
 
     /**
      * Adds all of the supplied data to the request data parameters
      * @param data collection of data parameters
      * @return this Connection, for chaining
      */
-    public Connection data(Collection<KeyVal> data);
+    Connection data(Collection<KeyVal> data);
 
     /**
      * Adds all of the supplied data to the request data parameters
      * @param data map of data parameters
      * @return this Connection, for chaining
      */
-    public Connection data(Map<String, String> data);
+    Connection data(Map<String, String> data);
 
     /**
-     * Add a number of request data parameters. Multiple parameters may be set at once, e.g.:
-     * <code>.data("name", "jsoup", "language", "Java", "language", "English");</code> creates a query string like:
-     * <code>?name=jsoup&language=Java&language=English</code>
+     * Add a number of request data parameters. Multiple parameters may be set at once, e.g.: <code>.data("name",
+     * "jsoup", "language", "Java", "language", "English");</code> creates a query string like:
+     * <code>{@literal ?name=jsoup&language=Java&language=English}</code>
      * @param keyvals a set of key value pairs.
      * @return this Connection, for chaining
      */
-    public Connection data(String... keyvals);
+    Connection data(String... keyvals);
 
     /**
      * Set a request header.
@@ -144,7 +193,7 @@ public interface Connection {
      * @return this Connection, for chaining
      * @see org.jsoup.Connection.Request#headers()
      */
-    public Connection header(String name, String value);
+    Connection header(String name, String value);
 
     /**
      * Set a cookie to be sent in the request.
@@ -152,21 +201,29 @@ public interface Connection {
      * @param value value of cookie
      * @return this Connection, for chaining
      */
-    public Connection cookie(String name, String value);
+    Connection cookie(String name, String value);
 
     /**
      * Adds each of the supplied cookies to the request.
-     * @param cookies map of cookie name -> value pairs
+     * @param cookies map of cookie name {@literal ->} value pairs
      * @return this Connection, for chaining
      */
-    public Connection cookies(Map<String, String> cookies);
+    Connection cookies(Map<String, String> cookies);
 
     /**
-     * Provide an alternate parser to use when parsing the response to a Document.
+     * Provide an alternate parser to use when parsing the response to a Document. If not set, defaults to the HTML
+     * parser, unless the response content-type is XML, in which case the XML parser is used.
      * @param parser alternate parser
      * @return this Connection, for chaining
      */
-    public Connection parser(Parser parser);
+    Connection parser(Parser parser);
+
+    /**
+     * Sets the default post data character set for x-www-form-urlencoded post data
+     * @param charset character set to encode post data
+     * @return this Connection, for chaining
+     */
+    Connection postDataCharset(String charset);
 
     /**
      * Execute the request as a GET, and parse the result.
@@ -177,7 +234,7 @@ public interface Connection {
      * @throws java.net.SocketTimeoutException if the connection times out
      * @throws IOException on error
      */
-    public Document get() throws IOException;
+    Document get() throws IOException;
 
     /**
      * Execute the request as a POST, and parse the result.
@@ -188,7 +245,7 @@ public interface Connection {
      * @throws java.net.SocketTimeoutException if the connection times out
      * @throws IOException on error
      */
-    public Document post() throws IOException;
+    Document post() throws IOException;
 
     /**
      * Execute the request.
@@ -199,34 +256,33 @@ public interface Connection {
      * @throws java.net.SocketTimeoutException if the connection times out
      * @throws IOException on error
      */
-    public Response execute() throws IOException;
+    Response execute() throws IOException;
 
     /**
      * Get the request object associated with this connection
      * @return request
      */
-    public Request request();
+    Request request();
 
     /**
      * Set the connection's request
      * @param request new request object
      * @return this Connection, for chaining
      */
-    public Connection request(Request request);
+    Connection request(Request request);
 
     /**
      * Get the response, once the request has been executed
      * @return response
      */
-    public Response response();
+    Response response();
 
     /**
      * Set the connection's response
      * @param response new response
      * @return this Connection, for chaining
      */
-    public Connection response(Response response);
-
+    Connection response(Response response);
 
     /**
      * Common methods for Requests and Responses
@@ -238,76 +294,86 @@ public interface Connection {
          * Get the URL
          * @return URL
          */
-        public URL url();
+        URL url();
 
         /**
          * Set the URL
          * @param url new URL
          * @return this, for chaining
          */
-        public T url(URL url);
+        T url(URL url);
 
         /**
          * Get the request method
          * @return method
          */
-        public Method method();
+        Method method();
 
         /**
          * Set the request method
          * @param method new method
          * @return this, for chaining
          */
-        public T method(Method method);
+        T method(Method method);
 
         /**
          * Get the value of a header. This is a simplified header model, where a header may only have one value.
          * <p>
          * Header names are case insensitive.
+         * </p>
          * @param name name of header (case insensitive)
          * @return value of header, or null if not set.
          * @see #hasHeader(String)
          * @see #cookie(String)
          */
-        public String header(String name);
+        String header(String name);
 
         /**
-         * Set a header. This method will overwrite any existing header with the same case insensitive name. 
+         * Set a header. This method will overwrite any existing header with the same case insensitive name.
          * @param name Name of header
          * @param value Value of header
          * @return this, for chaining
          */
-        public T header(String name, String value);
+        T header(String name, String value);
 
         /**
          * Check if a header is present
          * @param name name of header (case insensitive)
          * @return if the header is present in this request/response
          */
-        public boolean hasHeader(String name);
+        boolean hasHeader(String name);
+
+        /**
+         * Check if a header is present, with the given value
+         * @param name header name (case insensitive)
+         * @param value value (case insensitive)
+         * @return if the header and value pair are set in this req/res
+         */
+        boolean hasHeaderWithValue(String name, String value);
 
         /**
          * Remove a header by name
          * @param name name of header to remove (case insensitive)
          * @return this, for chaining
          */
-        public T removeHeader(String name);
+        T removeHeader(String name);
 
         /**
          * Retrieve all of the request/response headers as a map
          * @return headers
          */
-        public Map<String, String> headers();
+        Map<String, String> headers();
 
         /**
          * Get a cookie value by name from this request/response.
          * <p>
          * Response objects have a simplified cookie model. Each cookie set in the response is added to the response
          * object's cookie key=value map. The cookie's path, domain, and expiry date are ignored.
+         * </p>
          * @param name name of cookie to retrieve.
          * @return value of cookie, or null if not set
          */
-        public String cookie(String name);
+        String cookie(String name);
 
         /**
          * Set a cookie in this request/response.
@@ -315,206 +381,252 @@ public interface Connection {
          * @param value value of cookie
          * @return this, for chaining
          */
-        public T cookie(String name, String value);
+        T cookie(String name, String value);
 
         /**
          * Check if a cookie is present
          * @param name name of cookie
          * @return if the cookie is present in this request/response
          */
-        public boolean hasCookie(String name);
+        boolean hasCookie(String name);
 
         /**
          * Remove a cookie by name
          * @param name name of cookie to remove
          * @return this, for chaining
          */
-        public T removeCookie(String name);
+        T removeCookie(String name);
 
         /**
          * Retrieve all of the request/response cookies as a map
          * @return cookies
          */
-        public Map<String, String> cookies();
-
+        Map<String, String> cookies();
     }
 
     /**
      * Represents a HTTP request.
      */
-    public interface Request extends Base<Request> {
+    interface Request extends Base<Request> {
+
+
         /**
          * Get the request timeout, in milliseconds.
          * @return the timeout in milliseconds.
          */
-        public int timeout();
+        int timeout();
 
         /**
          * Update the request timeout.
          * @param millis timeout, in milliseconds
          * @return this Request, for chaining
          */
-        public Request timeout(int millis);
+        Request timeout(int millis);
 
         /**
          * Get the maximum body size, in bytes.
          * @return the maximum body size, in bytes.
          */
-        public int maxBodySize();
+        int maxBodySize();
 
         /**
          * Update the maximum body size, in bytes.
          * @param bytes maximum body size, in bytes.
          * @return this Request, for chaining
          */
-        public Request maxBodySize(int bytes);
+        Request maxBodySize(int bytes);
 
         /**
          * Get the current followRedirects configuration.
          * @return true if followRedirects is enabled.
          */
-        public boolean followRedirects();
+        boolean followRedirects();
 
         /**
          * Configures the request to (not) follow server redirects. By default this is <b>true</b>.
-         *
          * @param followRedirects true if server redirects should be followed.
          * @return this Request, for chaining
          */
-        public Request followRedirects(boolean followRedirects);
+        Request followRedirects(boolean followRedirects);
 
         /**
          * Get the current ignoreHttpErrors configuration.
-         * @return true if errors will be ignored; false (default) if HTTP errors will cause an IOException to be thrown.
+         * @return true if errors will be ignored; false (default) if HTTP errors will cause an IOException to be
+         * thrown.
          */
-        public boolean ignoreHttpErrors();
+        boolean ignoreHttpErrors();
 
-    	/**
-    	 * Configures the request to ignore HTTP errors in the response.
-    	 * @param ignoreHttpErrors set to true to ignore HTTP errors.
+        /**
+         * Configures the request to ignore HTTP errors in the response.
+         * @param ignoreHttpErrors set to true to ignore HTTP errors.
          * @return this Request, for chaining
-    	 */
-        public Request ignoreHttpErrors(boolean ignoreHttpErrors);
+         */
+        Request ignoreHttpErrors(boolean ignoreHttpErrors);
 
         /**
          * Get the current ignoreContentType configuration.
-         * @return true if invalid content-types will be ignored; false (default) if they will cause an IOException to be thrown.
+         * @return true if invalid content-types will be ignored; false (default) if they will cause an IOException to
+         * be thrown.
          */
-        public boolean ignoreContentType();
+        boolean ignoreContentType();
 
         /**
-    	 * Configures the request to ignore the Content-Type of the response.
-    	 * @param ignoreContentType set to true to ignore the content type.
+         * Configures the request to ignore the Content-Type of the response.
+         * @param ignoreContentType set to true to ignore the content type.
          * @return this Request, for chaining
-    	 */
-        public Request ignoreContentType(boolean ignoreContentType);
+         */
+        Request ignoreContentType(boolean ignoreContentType);
+
+        /**
+         * Get the current state of TLS (SSL) certificate validation.
+         * @return true if TLS cert validation enabled
+         */
+        boolean validateTLSCertificates();
+
+        /**
+         * Set TLS certificate validation.
+         * @param value set false to ignore TLS (SSL) certificates
+         */
+        void validateTLSCertificates(boolean value);
 
         /**
          * Add a data parameter to the request
          * @param keyval data to add.
          * @return this Request, for chaining
          */
-        public Request data(KeyVal keyval);
+        Request data(KeyVal keyval);
 
         /**
          * Get all of the request's data parameters
          * @return collection of keyvals
          */
-        public Collection<KeyVal> data();
+        Collection<KeyVal> data();
 
         /**
          * Specify the parser to use when parsing the document.
          * @param parser parser to use.
          * @return this Request, for chaining
          */
-        public Request parser(Parser parser);
+        Request parser(Parser parser);
 
         /**
          * Get the current parser to use when parsing the document.
          * @return current Parser
          */
-        public Parser parser();
+        Parser parser();
+
+        /**
+         * Sets the post data character set for x-www-form-urlencoded post data
+         * @param charset character set to encode post data
+         * @return this Request, for chaining
+         */
+        Request postDataCharset(String charset);
+
+        /**
+         * Gets the post data character set for x-www-form-urlencoded post data
+         * @return character set to encode post data
+         */
+        String postDataCharset();
+
     }
 
     /**
      * Represents a HTTP response.
      */
-    public interface Response extends Base<Response> {
-    	
-    	/**
+    interface Response extends Base<Response> {
+
+        /**
          * Get the status code of the response.
          * @return status code
          */
-        public int statusCode();
+        int statusCode();
 
         /**
          * Get the status message of the response.
          * @return status message
          */
-        public String statusMessage();
+        String statusMessage();
 
         /**
          * Get the character set name of the response.
          * @return character set name
          */
-        public String charset();
+        String charset();
 
         /**
          * Get the response content type (e.g. "text/html");
          * @return the response content type
          */
-        public String contentType();
+        String contentType();
 
         /**
          * Parse the body of the response as a Document.
          * @return a parsed Document
          * @throws IOException on error
          */
-        public Document parse() throws IOException;
+        Document parse() throws IOException;
 
         /**
          * Get the body of the response as a plain string.
          * @return body
          */
-        public String body();
+        String body();
 
         /**
          * Get the body of the response as an array of bytes.
          * @return body bytes
          */
-        public byte[] bodyAsBytes();
+        byte[] bodyAsBytes();
     }
 
     /**
      * A Key Value tuple.
      */
-    public interface KeyVal {
+    interface KeyVal {
 
         /**
          * Update the key of a keyval
          * @param key new key
          * @return this KeyVal, for chaining
          */
-        public KeyVal key(String key);
+        KeyVal key(String key);
 
         /**
          * Get the key of a keyval
          * @return the key
          */
-        public String key();
+        String key();
 
         /**
          * Update the value of a keyval
          * @param value the new value
          * @return this KeyVal, for chaining
          */
-        public KeyVal value(String value);
+        KeyVal value(String value);
 
         /**
          * Get the value of a keyval
          * @return the value
          */
-        public String value();
+        String value();
+
+        /**
+         * Add or update an input stream to this keyVal
+         * @param inputStream new input stream
+         * @return this KeyVal, for chaining
+         */
+        KeyVal inputStream(InputStream inputStream);
+
+        /**
+         * Get the input stream associated with this keyval, if any
+         * @return input stream if set, or null
+         */
+        InputStream inputStream();
+
+        /**
+         * Does this keyval have an input stream?
+         * @return true if this keyval does indeed have an input stream
+         */
+        boolean hasInputStream();
     }
 }
-
