@@ -18,6 +18,7 @@ public class Document extends Element {
     private OutputSettings outputSettings = new OutputSettings();
     private QuirksMode quirksMode = QuirksMode.noQuirks;
     private String location;
+    private boolean updateMetaCharset = false;
 
     /**
      Create a new, empty Document.
@@ -128,6 +129,8 @@ public class Document extends Element {
         normaliseStructure("head", htmlEl);
         normaliseStructure("body", htmlEl);
         
+        ensureMetaCharsetElement();
+        
         return this;
     }
 
@@ -206,6 +209,77 @@ public class Document extends Element {
     public String nodeName() {
         return "#document";
     }
+    
+    /**
+     * Sets the charset used in this document. This method is equivalent
+     * to {@link OutputSettings#charset(java.nio.charset.Charset)
+     * OutputSettings.charset(Charset)} but in addition it updates the
+     * charset / encoding element within the document.
+     * 
+     * <p>This enables
+     * {@link #updateMetaCharsetElement(boolean) meta charset update}.</p>
+     * 
+     * <p>If there's no element with charset / encoding information yet it will
+     * be created. Obsolete charset / encoding definitions are removed!</p>
+     * 
+     * <p><b>Elements used:</b></p>
+     * 
+     * <ul>
+     * <li><b>Html:</b> <i>&lt;meta charset="CHARSET"&gt;</i></li>
+     * <li><b>Xml:</b> <i>&lt;?xml version="1.0" encoding="CHARSET"&gt;</i></li>
+     * </ul>
+     * 
+     * @param charset Charset
+     * 
+     * @see #updateMetaCharsetElement(boolean) 
+     * @see OutputSettings#charset(java.nio.charset.Charset) 
+     */
+    public void charset(Charset charset) {
+        updateMetaCharsetElement(true);
+        outputSettings.charset(charset);
+        ensureMetaCharsetElement();
+    }
+    
+    /**
+     * Returns the charset used in this document. This method is equivalent
+     * to {@link OutputSettings#charset()}.
+     * 
+     * @return Current Charset
+     * 
+     * @see OutputSettings#charset() 
+     */
+    public Charset charset() {
+        return outputSettings.charset();
+    }
+    
+    /**
+     * Sets whether the element with charset information in this document is
+     * updated on changes through {@link #charset(java.nio.charset.Charset)
+     * Document.charset(Charset)} or not.
+     * 
+     * <p>If set to <tt>false</tt> <i>(default)</i> there are no elements
+     * modified.</p>
+     * 
+     * @param update If <tt>true</tt> the element updated on charset
+     * changes, <tt>false</tt> if not
+     * 
+     * @see #charset(java.nio.charset.Charset) 
+     */
+    public void updateMetaCharsetElement(boolean update) {
+        this.updateMetaCharset = update;
+    }
+    
+    /**
+     * Returns whether the element with charset information in this document is
+     * updated on changes through {@link #charset(java.nio.charset.Charset)
+     * Document.charset(Charset)} or not.
+     * 
+     * @return Returns <tt>true</tt> if the element is updated on charset
+     * changes, <tt>false</tt> if not
+     */
+    public boolean updateMetaCharsetElement() {
+        return updateMetaCharset;
+    }
 
     @Override
     public Document clone() {
@@ -213,6 +287,79 @@ public class Document extends Element {
         clone.outputSettings = this.outputSettings.clone();
         return clone;
     }
+    
+    /**
+     * Ensures a meta charset (html) or xml declaration (xml) with the current
+     * encoding used. This only applies with
+     * {@link #updateMetaCharsetElement(boolean) updateMetaCharset} set to
+     * <tt>true</tt>, otherwise this method does nothing.
+     * 
+     * <ul>
+     * <li>An exsiting element gets updated with the current charset</li>
+     * <li>If there's no element yet it will be inserted</li>
+     * <li>Obsolete elements are removed</li>
+     * </ul>
+     * 
+     * <p><b>Elements used:</b></p>
+     * 
+     * <ul>
+     * <li><b>Html:</b> <i>&lt;meta charset="CHARSET"&gt;</i></li>
+     * <li><b>Xml:</b> <i>&lt;?xml version="1.0" encoding="CHARSET"&gt;</i></li>
+     * </ul>
+     */
+    private void ensureMetaCharsetElement() {
+        if (updateMetaCharset) {
+            OutputSettings.Syntax syntax = outputSettings().syntax();
+
+            if (syntax == OutputSettings.Syntax.html) {
+                Element metaCharset = select("meta[charset]").first();
+
+                if (metaCharset != null) {
+                    metaCharset.attr("charset", charset().displayName());
+                } else {
+                    Element head = head();
+
+                    if (head != null) {
+                        head.appendElement("meta").attr("charset", charset().displayName());
+                    }
+                }
+
+                // Remove obsolete elements
+                select("meta[name=charset]").remove();
+            } else if (syntax == OutputSettings.Syntax.xml) {
+                Node node = childNodes().get(0);
+
+                if (node instanceof XmlDeclaration) {
+                    XmlDeclaration decl = (XmlDeclaration) node;
+
+                    if (decl.attr(XmlDeclaration.DECL_KEY).equals("xml")) {
+                        decl.attr("encoding", charset().displayName());
+
+                        final String version = decl.attr("version");
+
+                        if (version != null) {
+                            decl.attr("version", "1.0");
+                        }
+                    } else {
+                        decl = new XmlDeclaration("xml", baseUri, false);
+                        decl.attr("version", "1.0");
+                        decl.attr("encoding", charset().displayName());
+
+                        prependChild(decl);
+                    }
+                } else {
+                    XmlDeclaration decl = new XmlDeclaration("xml", baseUri, false);
+                    decl.attr("version", "1.0");
+                    decl.attr("encoding", charset().displayName());
+
+                    prependChild(decl);
+                }
+            } else {
+                // Unsupported syntax - nothing to do yet
+            }
+        }
+    }
+    
 
     /**
      * A Document's output settings control the form of the text() and html() methods.
@@ -232,7 +379,7 @@ public class Document extends Element {
         private Syntax syntax = Syntax.html;
 
         public OutputSettings() {}
-
+        
         /**
          * Get the document's current HTML escape mode: <code>base</code>, which provides a limited set of named HTML
          * entities and escapes other characters as numbered entities for maximum compatibility; or <code>extended</code>,
@@ -274,7 +421,6 @@ public class Document extends Element {
          * @return the document's output settings, for chaining
          */
         public OutputSettings charset(Charset charset) {
-            // todo: this should probably update the doc's meta charset
             this.charset = charset;
             charsetEncoder = charset.newEncoder();
             return this;
@@ -361,7 +507,7 @@ public class Document extends Element {
 
         /**
          * Set the indent amount for pretty printing
-         * @param indentAmount number of spaces to use for indenting each level. Must be >= 0.
+         * @param indentAmount number of spaces to use for indenting each level. Must be {@literal >=} 0.
          * @return this, for chaining
          */
         public OutputSettings indentAmount(int indentAmount) {
@@ -405,7 +551,7 @@ public class Document extends Element {
     }
 
     public enum QuirksMode {
-        noQuirks, quirks, limitedQuirks;
+        noQuirks, quirks, limitedQuirks
     }
 
     public QuirksMode quirksMode() {
@@ -416,10 +562,4 @@ public class Document extends Element {
         this.quirksMode = quirksMode;
         return this;
     }
-
-    @Override
-    public boolean equals(Object o) {
-        return super.equals(o);
-    }
 }
-
