@@ -7,8 +7,6 @@ import java.io.IOException;
 import java.io.InputStream;
 import java.nio.charset.CharsetEncoder;
 import java.util.*;
-import java.util.regex.Matcher;
-import java.util.regex.Pattern;
 
 /**
  * HTML entities, and escape routines.
@@ -65,7 +63,7 @@ public class Entities {
     /**
      * Get the Character value of the named entity
      * @param name named entity (e.g. "lt" or "amp")
-     * @return the Character value of the named entity (e.g. '<' or '&')
+     * @return the Character value of the named entity (e.g. '{@literal <}' or '{@literal &}')
      */
     public static Character getCharacterByName(String name) {
         return full.get(name);
@@ -83,9 +81,10 @@ public class Entities {
 
         boolean lastWasWhite = false;
         boolean reachedNonWhite = false;
-        EscapeMode escapeMode = out.escapeMode();
-        CharsetEncoder encoder = out.encoder();
-        Map<Character, String> map = escapeMode.getMap();
+        final EscapeMode escapeMode = out.escapeMode();
+        final CharsetEncoder encoder = out.encoder();
+        final CoreCharset coreCharset = CoreCharset.byName(encoder.charset().name());
+        final Map<Character, String> map = escapeMode.getMap();
         final int length = string.length();
 
         int codePoint;
@@ -116,10 +115,11 @@ public class Entities {
                         if (escapeMode != EscapeMode.xhtml)
                             accum.append("&nbsp;");
                         else
-                            accum.append(c);
+                            accum.append("&#xa0;");
                         break;
                     case '<':
-                        if (!inAttribute)
+                        // escape when in character data or when in a xml attribue val; not needed in html attr val
+                        if (!inAttribute || escapeMode == EscapeMode.xhtml)
                             accum.append("&lt;");
                         else
                             accum.append(c);
@@ -137,7 +137,7 @@ public class Entities {
                             accum.append(c);
                         break;
                     default:
-                        if (encoder.canEncode(c))
+                        if (canEncode(coreCharset, c, encoder))
                             accum.append(c);
                         else if (map.containsKey(c))
                             accum.append('&').append(map.get(c)).append(';');
@@ -146,7 +146,7 @@ public class Entities {
                 }
             } else {
                 final String c = new String(Character.toChars(codePoint));
-                if (encoder.canEncode(c))
+                if (encoder.canEncode(c)) // uses fallback encoder for simplicity
                     accum.append(c);
                 else
                     accum.append("&#x").append(Integer.toHexString(codePoint)).append(';');
@@ -160,14 +160,52 @@ public class Entities {
 
     /**
      * Unescape the input string.
-     * @param string
+     * @param string to un-HTML-escape
      * @param strict if "strict" (that is, requires trailing ';' char, otherwise that's optional)
-     * @return
+     * @return unescaped string
      */
     static String unescape(String string, boolean strict) {
         return Parser.unescapeEntities(string, strict);
     }
 
+    /*
+     * Provides a fast-path for Encoder.canEncode, which drastically improves performance on Android post JellyBean.
+     * After KitKat, the implementation of canEncode degrades to the point of being useless. For non ASCII or UTF,
+     * performance may be bad. We can add more encoders for common character sets that are impacted by performance
+     * issues on Android if required.
+     *
+     * Benchmarks:     *
+     * OLD toHtml() impl v New (fastpath) in millis
+     * Wiki: 1895, 16
+     * CNN: 6378, 55
+     * Alterslash: 3013, 28
+     * Jsoup: 167, 2
+     */
+
+    private static boolean canEncode(final CoreCharset charset, final char c, final CharsetEncoder fallback) {
+        // todo add more charset tests if impacted by Android's bad perf in canEncode
+        switch (charset) {
+            case ascii:
+                return c < 0x80;
+            case utf:
+                return true; // real is:!(Character.isLowSurrogate(c) || Character.isHighSurrogate(c)); - but already check above
+            default:
+                return fallback.canEncode(c);
+        }
+    }
+
+    private enum CoreCharset {
+        ascii, utf, fallback;
+
+        private static CoreCharset byName(String name) {
+            if (name.equals("US-ASCII"))
+                return ascii;
+            if (name.startsWith("UTF-")) // covers UTF-8, UTF-16, et al
+                return utf;
+            return fallback;
+        }
+    }
+
 
     // xhtml has restricted entities
     private static final Object[][] xhtmlArray = {
