@@ -5,7 +5,7 @@ package org.jsoup.nodes;
 
  @author Jonathan Hedley, jonathan@hedley.net */
 public class XmlDeclaration extends Node {
-    private static final String DECL_KEY = "declaration";
+    static final String DECL_KEY = "declaration";
     private final boolean isProcessingInstruction; // <! if true, <? if false, declaration (and last data char should be ?)
 
     /**
@@ -29,9 +29,29 @@ public class XmlDeclaration extends Node {
      @return XML declaration
      */
     public String getWholeDeclaration() {
-        return attributes.get(DECL_KEY);
+        final String decl = attributes.get(DECL_KEY);
+        
+        if(decl.equals("xml") && attributes.size() > 1 ) {
+            StringBuilder sb = new StringBuilder(decl);
+            final String version = attributes.get("version");
+            
+            if( version != null ) {
+                sb.append(" version=\"").append(version).append("\"");
+            }
+            
+            final String encoding = attributes.get("encoding");
+            
+            if( encoding != null ) {
+                sb.append(" encoding=\"").append(encoding).append("\"");
+            }
+            
+            return sb.toString();
+        }
+        else {
+            return attributes.get(DECL_KEY);
+        }
     }
-
+    
     void outerHtmlHead(StringBuilder accum, int depth, Document.OutputSettings out) {
         accum
                 .append("<")
@@ -42,6 +62,7 @@ public class XmlDeclaration extends Node {
 
     void outerHtmlTail(StringBuilder accum, int depth, Document.OutputSettings out) {}
 
+    @Override
     public String toString() {
         return outerHtml();
     }
