@@ -209,8 +209,7 @@ public class TokenQueue {
                 pos++;
         }
 
-        String data = queue.substring(start, pos); 
-        return data; 
+        return queue.substring(start, pos);
     }
 
     /**
@@ -226,8 +225,7 @@ public class TokenQueue {
             pos++;
         }
 
-        String data = queue.substring(start, pos); 
-        return data; 
+        return queue.substring(start, pos);
     }
 
     /**
@@ -307,6 +305,7 @@ public class TokenQueue {
 
     /**
      * Pulls the next run of whitespace characters of the queue.
+     * @return Whether consuming whitespace or not
      */
     public boolean consumeWhitespace() {
         boolean seen = false;
@@ -389,6 +388,7 @@ public class TokenQueue {
         return remainder;
     }
     
+    @Override
     public String toString() {
         return queue.substring(pos);
     }
