@@ -25,6 +25,8 @@ public abstract class Evaluator {
      *
      * @param root    Root of the matching subtree
      * @param element tested element
+     * @return Returns <tt>true</tt> if the requirements are met or
+     * <tt>false</tt> otherwise
      */
     public abstract boolean matches(Element root, Element element);
 
@@ -152,7 +154,7 @@ public abstract class Evaluator {
 
         @Override
         public boolean matches(Element root, Element element) {
-            return element.hasAttr(key) && value.equalsIgnoreCase(element.attr(key));
+            return element.hasAttr(key) && value.equalsIgnoreCase(element.attr(key).trim());
         }
 
         @Override
@@ -302,7 +304,7 @@ public abstract class Evaluator {
     }
 
     /**
-     * Evaluator for matching by sibling index number (e < idx)
+     * Evaluator for matching by sibling index number (e {@literal <} idx)
      */
     public static final class IndexLessThan extends IndexEvaluator {
         public IndexLessThan(int index) {
@@ -322,7 +324,7 @@ public abstract class Evaluator {
     }
 
     /**
-     * Evaluator for matching by sibling index number (e > idx)
+     * Evaluator for matching by sibling index number (e {@literal >} idx)
      */
     public static final class IndexGreaterThan extends IndexEvaluator {
         public IndexGreaterThan(int index) {
