@@ -4,17 +4,25 @@ import org.jsoup.Connection;
 import org.jsoup.HttpStatusException;
 import org.jsoup.Jsoup;
 import org.jsoup.UnsupportedMimeTypeException;
+import org.jsoup.helper.StringUtil;
+import org.jsoup.helper.W3CDom;
 import org.jsoup.nodes.Document;
+import org.jsoup.nodes.FormElement;
+import org.jsoup.parser.HtmlTreeBuilder;
+import org.jsoup.parser.Parser;
+import org.jsoup.parser.XmlTreeBuilder;
 import org.junit.Ignore;
 import org.junit.Test;
 
+import java.io.File;
+import java.io.FileInputStream;
 import java.io.IOException;
 import java.net.MalformedURLException;
 import java.net.URL;
 import java.util.Map;
 
-import static org.junit.Assert.assertEquals;
-import static org.junit.Assert.assertTrue;
+import static org.hamcrest.core.Is.is;
+import static org.junit.Assert.*;
 
 /**
  Tests the URL connection. Not enabled by default, so tests don't require network connection.
@@ -22,13 +30,15 @@ import static org.junit.Assert.assertTrue;
  @author Jonathan Hedley, jonathan@hedley.net */
 @Ignore // ignored by default so tests don't require network access. comment out to enable.
 public class UrlConnectTest {
+    private static final String WEBSITE_WITH_INVALID_CERTIFICATE = "https://certs.cac.washington.edu/CAtest/";
+    private static final String WEBSITE_WITH_SNI = "https://jsoup.org/";
     private static String echoURL = "http://direct.infohound.net/tools/q.pl";
 
     @Test
     public void fetchURl() throws IOException {
-        String url = "http://www.google.com"; // no trailing / to force redir
+        String url = "http://jsoup.org"; // no trailing / to force redir
         Document doc = Jsoup.parse(new URL(url), 10*1000);
-        assertTrue(doc.title().contains("Google"));
+        assertTrue(doc.title().contains("jsoup"));
     }
 
     @Test
@@ -51,13 +61,13 @@ public class UrlConnectTest {
     
     @Test
     public void exceptOnUnknownContentType() {
-        String url = "http://jsoup.org/rez/osi_logo.png"; // not text/* but image/png, should throw
+        String url = "http://direct.jsoup.org/rez/osi_logo.png"; // not text/* but image/png, should throw
         boolean threw = false;
         try {
             Document doc = Jsoup.parse(new URL(url), 3000);
         } catch (UnsupportedMimeTypeException e) {
             threw = true;
-            assertEquals("org.jsoup.UnsupportedMimeTypeException: Unhandled content type. Must be text/*, application/xml, or application/xhtml+xml. Mimetype=image/png, URL=http://jsoup.org/rez/osi_logo.png", e.toString());
+            assertEquals("org.jsoup.UnsupportedMimeTypeException: Unhandled content type. Must be text/*, application/xml, or application/xhtml+xml. Mimetype=image/png, URL=http://direct.jsoup.org/rez/osi_logo.png", e.toString());
             assertEquals(url, e.getUrl());
             assertEquals("image/png", e.getMimeType());
         } catch (IOException e) {
@@ -113,6 +123,21 @@ public class UrlConnectTest {
         assertEquals("http://example.com", ihVal("HTTP_REFERER", doc));
     }
 
+    @Test
+    public void doesPut() throws IOException {
+        Connection.Response res = Jsoup.connect(echoURL)
+                .data("uname", "Jsoup", "uname", "Jonathan", "百", "度一下")
+                .cookie("auth", "token")
+                .method(Connection.Method.PUT)
+                .execute();
+
+        Document doc = res.parse();
+        assertEquals("PUT", ihVal("REQUEST_METHOD", doc));
+        //assertEquals("gzip", ihVal("HTTP_ACCEPT_ENCODING", doc)); // current proxy removes gzip on post
+        assertEquals("auth=token", ihVal("HTTP_COOKIE", doc));
+    }
+
+
     private static String ihVal(String key, Document doc) {
         return doc.select("th:contains("+key+") + td").first().text();
     }
@@ -152,11 +177,29 @@ public class UrlConnectTest {
 
     @Test
     public void followsRelativeRedirect() throws IOException {
-        Connection con = Jsoup.connect("http://direct.infohound.net/tools/302-rel.pl"); // to ./ - /tools/
+        Connection con = Jsoup.connect("http://direct.infohound.net/tools/302-rel.pl"); // to /tidy/
         Document doc = con.post();
         assertTrue(doc.title().contains("HTML Tidy Online"));
     }
 
+    @Test
+    public void followsRelativeDotRedirect() throws IOException {
+        // redirects to "./ok.html", should resolve to http://direct.infohound.net/tools/ok.html
+        Connection con = Jsoup.connect("http://direct.infohound.net/tools/302-rel-dot.pl"); // to ./ok.html
+        Document doc = con.post();
+        assertTrue(doc.title().contains("OK"));
+        assertEquals(doc.location(), "http://direct.infohound.net/tools/ok.html");
+    }
+
+    @Test
+    public void followsRelativeDotRedirect2() throws IOException {
+        //redirects to "esportspenedes.cat/./ep/index.php", should resolve to "esportspenedes.cat/ep/index.php"
+        Connection con = Jsoup.connect("http://esportspenedes.cat")  // note lack of trailing / - server should redir to / first, then to ./ep/...; but doesn't'
+                .timeout(10000);
+        Document doc = con.post();
+        assertEquals(doc.location(), "http://esportspenedes.cat/ep/index.php");
+    }
+
     @Test
     public void followsRedirectsWithWithespaces() throws IOException {
         Connection con = Jsoup.connect("http://tinyurl.com/kgofxl8"); // to http://www.google.com/?q=white spaces
@@ -197,6 +240,34 @@ public class UrlConnectTest {
         assertEquals("404 Not Found", doc.select("h1").first().text());
     }
 
+    @Test
+    public void ignores500tExceptionIfSoConfigured() throws IOException {
+        Connection con = Jsoup.connect("http://direct.infohound.net/tools/500.pl").ignoreHttpErrors(true);
+        Connection.Response res = con.execute();
+        Document doc = res.parse();
+        assertEquals(500, res.statusCode());
+        assertEquals("Application Error", res.statusMessage());
+        assertEquals("Woops", doc.select("h1").first().text());
+    }
+
+    @Test
+    public void ignores500NoWithContentExceptionIfSoConfigured() throws IOException {
+        Connection con = Jsoup.connect("http://direct.infohound.net/tools/500-no-content.pl").ignoreHttpErrors(true);
+        Connection.Response res = con.execute();
+        Document doc = res.parse();
+        assertEquals(500, res.statusCode());
+        assertEquals("Application Error", res.statusMessage());
+    }
+
+    @Test
+    public void ignores200NoWithContentExceptionIfSoConfigured() throws IOException {
+        Connection con = Jsoup.connect("http://direct.infohound.net/tools/200-no-content.pl").ignoreHttpErrors(true);
+        Connection.Response res = con.execute();
+        Document doc = res.parse();
+        assertEquals(200, res.statusCode());
+        assertEquals("All Good", res.statusMessage());
+    }
+
     @Test
     public void doesntRedirectIfSoConfigured() throws IOException {
         Connection con = Jsoup.connect("http://direct.infohound.net/tools/302.pl").followRedirects(false);
@@ -278,6 +349,63 @@ public class UrlConnectTest {
         assertEquals(actualDocText, unlimitedRes.parse().text().length());
     }
 
+    /**
+     * Verify that security disabling feature works properly.
+     * <p/>
+     * 1. try to hit url with invalid certificate and evaluate that exception is thrown
+     *
+     * @throws Exception
+     */
+    @Test(expected = IOException.class)
+    public void testUnsafeFail() throws Exception {
+        String url = WEBSITE_WITH_INVALID_CERTIFICATE;
+        Jsoup.connect(url).execute();
+    }
+
+
+    /**
+     * Verify that requests to websites with SNI fail on jdk 1.6
+     * <p/>
+     * read for more details:
+     * http://en.wikipedia.org/wiki/Server_Name_Indication
+     *
+     * Test is ignored independent from others as it requires JDK 1.6
+     * @throws Exception
+     */
+    @Test(expected = IOException.class)
+    public void testSNIFail() throws Exception {
+        String url = WEBSITE_WITH_SNI;
+        Jsoup.connect(url).execute();
+    }
+
+    /**
+     * Verify that requests to websites with SNI pass
+     * <p/>
+     * <b>NB!</b> this test is FAILING right now on jdk 1.6
+     *
+     * @throws Exception
+     */
+    @Test
+    public void testSNIPass() throws Exception {
+        String url = WEBSITE_WITH_SNI;
+        Connection.Response defaultRes = Jsoup.connect(url).validateTLSCertificates(false).execute();
+        assertThat(defaultRes.statusCode(), is(200));
+    }
+
+    /**
+     * Verify that security disabling feature works properly.
+     * <p/>
+     * 1. disable security checks and call the same url to verify that content is consumed correctly
+     *
+     * @throws Exception
+     */
+    @Test
+    public void testUnsafePass() throws Exception {
+        String url = WEBSITE_WITH_INVALID_CERTIFICATE;
+        Connection.Response defaultRes = Jsoup.connect(url).validateTLSCertificates(false).execute();
+        assertThat(defaultRes.statusCode(), is(200));
+    }
+
     @Test
     public void shouldWorkForCharsetInExtraAttribute() throws IOException {
         Connection.Response res = Jsoup.connect("https://www.creditmutuel.com/groupe/fr/").execute();
@@ -325,4 +453,89 @@ public class UrlConnectTest {
         assertEquals("http://example.com/foo.jpg", doc.select("img").first().absUrl("src"));
     }
 
+    /**
+     * Test fetching a form, and submitting it with a file attached.
+     */
+    @Test
+    public void postHtmlFile() throws IOException {
+        Document index = Jsoup.connect("http://direct.infohound.net/tidy/").get();
+        FormElement form = index.select("[name=tidy]").forms().get(0);
+        Connection post = form.submit();
+
+        File uploadFile = ParseTest.getFile("/htmltests/google-ipod.html");
+        FileInputStream stream = new FileInputStream(uploadFile);
+
+        // todo: need to add a better way to get an existing data field
+        for (Connection.KeyVal keyVal : post.request().data()) {
+            if (keyVal.key().equals("_file")) {
+                keyVal.value("check.html");
+                keyVal.inputStream(stream);
+            }
+        }
+
+        Connection.Response res;
+        try {
+            res = post.execute();
+        } finally {
+            stream.close();
+        }
+
+        Document out = res.parse();
+        assertTrue(out.text().contains("HTML Tidy Complete"));
+    }
+
+    /**
+     * Tests upload of binary content to a remote service.
+     */
+    @Test
+    public void postJpeg() throws IOException {
+        File thumb = ParseTest.getFile("/htmltests/thumb.jpg");
+        Document result = Jsoup
+                .connect("http://regex.info/exif.cgi")
+                .data("f", thumb.getName(), new FileInputStream(thumb))
+                .post();
+
+        assertEquals("Baseline DCT, Huffman coding", result.select("td:contains(Process) + td").text());
+    }
+
+    @Test
+    public void handles201Created() throws IOException {
+        Document doc = Jsoup.connect("http://direct.infohound.net/tools/201.pl").get(); // 201, location=jsoup
+        assertEquals("http://jsoup.org", doc.location());
+    }
+
+    @Test
+    public void fetchToW3c() throws IOException {
+        String url = "http://jsoup.org";
+        Document doc = Jsoup.connect(url).get();
+
+        W3CDom dom = new W3CDom();
+        org.w3c.dom.Document wDoc = dom.fromJsoup(doc);
+        assertEquals(url, wDoc.getDocumentURI());
+        String html = dom.asString(wDoc);
+        assertTrue(html.contains("jsoup"));
+    }
+
+    @Test
+    public void fetchHandlesXml() throws IOException {
+        // should auto-detect xml and use XML parser, unless explicitly requested the html parser
+        String xmlUrl = "http://direct.infohound.net/tools/parse-xml.xml";
+        Connection con = Jsoup.connect(xmlUrl);
+        Document doc = con.get();
+        Connection.Request req = con.request();
+        assertTrue(req.parser().getTreeBuilder() instanceof XmlTreeBuilder);
+        assertEquals("<xml> <link> one </link> <table> Two </table> </xml>", StringUtil.normaliseWhitespace(doc.outerHtml()));
+    }
+
+    @Test
+    public void fetchHandlesXmlAsHtmlWhenParserSet() throws IOException {
+        // should auto-detect xml and use XML parser, unless explicitly requested the html parser
+        String xmlUrl = "http://direct.infohound.net/tools/parse-xml.xml";
+        Connection con = Jsoup.connect(xmlUrl).parser(Parser.htmlParser());
+        Document doc = con.get();
+        Connection.Request req = con.request();
+        assertTrue(req.parser().getTreeBuilder() instanceof HtmlTreeBuilder);
+        assertEquals("<html> <head></head> <body> <xml> <link>one <table> Two </table> </xml> </body> </html>", StringUtil.normaliseWhitespace(doc.outerHtml()));
+    }
+
 }
